@@ -1,8 +1,8 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # coding=utf-8
 
-from __future__ import print_function
 import distutils.dir_util as distutils
+import contextlib
 import filecmp
 import hashlib
 import glob
@@ -12,8 +12,6 @@ import os.path
 import shutil
 from zipfile import ZipFile, ZIP_DEFLATED
 
-import biplist
-
 
 # Create parent directories for the given path if they don't exist
 def create_parent_dirs(path):
@@ -25,16 +23,23 @@ def create_parent_dirs(path):
             pass
 
 
+# Read a .plist file from the given path and return a dictionary representing
+# the contents
+def read_plist_from_path(plist_path):
+    with open(plist_path, 'rb') as plist_file:
+        return plistlib.load(plist_file)
+
+
 # Retrieve correct path to directory containing Alfred's user preferences
 def get_user_prefs_dir():
 
     library_dir = os.path.join(os.path.expanduser('~'), 'Library')
     try:
-        core_prefs = biplist.readPlist(os.path.join(
+        core_prefs = read_plist_from_path(os.path.join(
             library_dir, 'Preferences',
             'com.runningwithcrayons.Alfred-Preferences-3.plist'))
     except IOError:
-        core_prefs = biplist.readPlist(os.path.join(
+        core_prefs = read_plist_from_path(os.path.join(
             library_dir, 'Preferences',
             'com.runningwithcrayons.Alfred-Preferences.plist'))
 
@@ -56,11 +61,7 @@ def get_installed_workflow(workflow_bundle_id):
     # Find workflow whose bundle ID matches this workflow's
     for workflow_dir in workflow_dirs:
         info_path = os.path.join(workflow_dir, 'info.plist')
-        if hasattr(plistlib, 'load'):
-            with open(info_path, 'rb') as info_file:
-                info = plistlib.load(info_file)
-        else:
-            info = plistlib.readPlist(info_path)
+        info = read_plist_from_path(info_path)
         if info['bundleid'] == workflow_bundle_id:
             return workflow_dir, info
 
@@ -68,6 +69,17 @@ def get_installed_workflow(workflow_bundle_id):
     raise OSError('Workflow is not installed locally')
 
 
+# Retrieve the octal file permissions for the given file as a base-10 integer
+def get_permissions(file_path):
+    return os.stat(file_path).st_mode
+
+
+# Return True if the permissions of the two files are equal; return False
+# otherwise
+def cmp_permissions(src_file_path, dest_file_path):
+    return get_permissions(src_file_path) == get_permissions(dest_file_path)
+
+
 # Return True if the item counts for the given directories match; otherwise,
 # return False
 def check_dir_item_count_match(dir_path, dest_dir_path, dirs_cmp):
@@ -112,7 +124,8 @@ def dirs_are_equal(dir_path, dest_dir_path):
     return True
 
 
-# Check if resource (file or directory) is equal to destination resource
+# Return True if the resource (file or directory) is equal to the destination
+# resource; return False otherwise
 def resources_are_equal(resource_path, dest_resource_path):
 
     try:
@@ -120,31 +133,48 @@ def resources_are_equal(resource_path, dest_resource_path):
     except OSError:
         # Compare files if they are not directories
         try:
-            return filecmp.cmp(resource_path, dest_resource_path)
+            return (
+                filecmp.cmp(resource_path, dest_resource_path)
+                and
+                cmp_permissions(resource_path, dest_resource_path)
+            )
         except OSError:
             # Resources are not equal if either does not exist
             return False
 
 
 # Copy package resource to corresponding destination path
-def copy_resource(resource_path, dest_resource_path):
+def copy_resource(resource_path, dest_resource_path, force=False):
 
-    if not resources_are_equal(resource_path, dest_resource_path):
+    if force or not resources_are_equal(resource_path, dest_resource_path):
         try:
             distutils.copy_tree(resource_path, dest_resource_path)
         except distutils.DistutilsFileError:
+            with contextlib.suppress(FileNotFoundError):
+                os.remove(dest_resource_path)
             shutil.copy(resource_path, dest_resource_path)
         print('Copied {}'.format(resource_path))
+        return True
+    else:
+        return False
 
 
 # Copy all package resources to installed workflow
-def copy_pkg_resources(workflow_path, workflow_resources):
+def copy_pkg_resources(workflow_path, workflow_resources, force=False):
 
+    copied_any = False
     for resource_patt in workflow_resources:
         for resource_path in glob.iglob(resource_patt):
             create_parent_dirs(os.path.join(workflow_path, resource_path))
             dest_resource_path = os.path.join(workflow_path, resource_path)
-            copy_resource(resource_path, dest_resource_path)
+            copied = copy_resource(
+                resource_path,
+                dest_resource_path,
+                force=force)
+            if copied:
+                copied_any = True
+    if not copied_any:
+        print('Nothing to copy; workflow is already up-to-date')
 
 
 # Update the workflow README with the current project README
@@ -198,20 +228,17 @@ def export_workflow(workflow_path, archive_path):
 
 # Package installed workflow by copying resources from project, updating
 # README, and optionally exporting workflow
-def package_workflow(config, version, export_file):
+def package_workflow(config, version, export_file, force=False):
 
     workflow_path, info = get_installed_workflow(config['bundle_id'])
 
-    copy_pkg_resources(workflow_path, config['resources'])
+    copy_pkg_resources(workflow_path, config['resources'], force=force)
     if 'readme' in config:
         update_workflow_readme(info, config['readme'])
     update_workflow_version(info, version)
     plist_path = os.path.join(workflow_path, 'info.plist')
-    if hasattr(plistlib, 'dump'):
-        with open(plist_path, 'rb+') as plist_file:
-            plistlib.dump(info, plist_file)
-    else:
-        plistlib.writePlist(info, plist_path)
+    with open(plist_path, 'rb+') as plist_file:
+        plistlib.dump(info, plist_file)
 
     if export_file == '':
         export_file = config['export_file']
